(function(window, serial_scheme){

    // construct function
    function serial_port(serial){ this.__init(serial); }

    /**
     * 初始化对象
     * @param object serial 串口对象
     * @return this
     */
    serial_port.prototype.__init = function(serial){ serial_port.prototype.serial = serial; };

    // 本次指令详细信息
    serial_port.prototype.codes_info = {};

    // 尝试错误次数
    serial_port.prototype.max_retry  = 5;

    // 一个指令发送的次数
    serial_port.prototype.send_num    = 0;

    // 已经发生的接收错误次数
    serial_port.prototype.receive_err_num = 0;

    // 串口发送数据成功回调函数
    serial_port.prototype.send_succcess_context = '';

    // 口发送数据失败回调函数
    serial_port.prototype.send_error_context = '';

    // 串口方案
    serial_port.prototype.scheme = new serial_scheme();

    // 串口类实例
    serial_port.prototype.serial = '';

    // 锁
    serial_port.prototype.lock   = false;

    // 重发定时器id
    serial_port.prototype.send_time_id = '';

    // 长时间锁未释放定时器id
    serial_port.prototype.send_lock_id = '';

    /*
     * 获取设备串口列表
     * @param callback call 回调函数
     * @return mixed
     */
    serial_port.prototype.examine = function(call){ serial_port.prototype.serial.init(function(ret, err){ call && call(ret, err); }); }

    /*
     * 打开串口
     * @param callback call 回调函数
     * @return mixed
     */
    serial_port.prototype.openRead = function(call){
         var self = serial_port.prototype;

         self.serial.open({port: '/dev/ttyS4', baudRate: '9600'}, function(ret, err){ return call && call(ret, err); });
    }

    /*
     * 向串口发送数据
     * @param mixed data 向串口推送的数据
     * @param callback success 成功回调函数
     * @param callback error 达到最大尝试次数回调函数
     * @return mixed
     */
    serial_port.prototype.sendPort = function(data, success, error){
        var self = serial_port.prototype;

        if(self.lock){ return error && error('', {msg: '资源已锁,请等待锁释放', type: 1}); }

        var params = {sendType: 'hex', data: self.scheme.join_scheme(data.code)};

        clearTimeout(self.send_time_id);
        clearTimeout(self.send_lock_id);

        self.serial.send(params, function(ret, err){
            self.send_num++;
            if(ret.status == true && self.send_num <= self.max_retry){
                self.send_num = 0;

                if(data.is_echoes === true) // 某种类型的指令不会返回数据（is_echoes = false），直接忽略
                {
                    self.lock       = true;
                    self.codes_info = data;

                    self.send_time_id = setTimeout(self.sendPort(self.codes_info, self.send_error_context, self.send_succcess_context), 1000);

                    self.send_lock_id = setTimeout(function(){
                        self.lock = false;
                        return error && error('', {msg: '锁超时', type: 2});
                    }, 5000);

                    // 保存成功及失败上下文供发生异常时调用
                    self.send_error_context    = error;
                    self.send_succcess_context = success;
                }

                success && success(ret, err);
            }else{
                if(self.send_num >= self.max_retry){
                    self.lock = false;
                    self.send_num = 0;
                    return error && error(ret, err);
                }else{
                    return self.sendPort(data, success, error);
                }
            }
        });
    }

    /**
     * 接收串口数据
     * @param callback success 成功回调函数
     * @param callback error 达到最大尝试次数回调函数
     * @return mixed
     */
    serial_port.prototype.receive = function(success, error){
        var self = serial_port.prototype;

        self.serial.addEventListener({
  			    sendType: 'hex'
  			},function(ret, err){
            if(ret && ret.evenType == 'onDataReceived'){
                var code = self.scheme.check_data(ret.hasOwnProperty('data') ? ret.data : ''); // 读取重量、校秤跳过验证。因为不需要验证返回值
                var id   = isNaN(parseInt(self.codes_info.id)) ? 0 : parseInt(self.codes_info.id);

                if(code === false){
                    if(id >= 101 && id <= 10300){
                        self.receive_err_num++;

                        if(self.receive_err_num >= self.max_retry){
                            self.receive_err_num = 0;

                            self.lock = false;
                            return error && error(ret, err, self.codes_info);
                        }else{
                            self.lock = false;
                            return self.sendPort(self.codes_info, self.send_succcess_context, self.send_error_context);
                        }
                        return;
                    }
                    self.lock = false;
                    return error && error(ret, err, self.codes_info);
                }

                var response = self.scheme.parse_code();
                if(code === true && (self.codes_info.id == response.codes.id))
                {
                    var codes_info = '';
                    for(var i in self.codes_info){
                        codes_info[i] = self.codes_info[i];
                    }

                    self.lock = false;
                    self.codes_info = '';
                    self.receive_err_num = 0;

                    return success(ret, err, response, codes_info);
                }
            }else if(ret && ret.evenType == 'onDataSent'){
                clearTimeout(self.send_time_id);
                clearTimeout(self.send_lock_id);
            }
  			});
    }

    window.serial_port = serial_port; // 暴露到window全局作用域。注意这里没有暴露为serial、serialPort因为在高版本中这是系统提供的Api

})(window, serial_scheme);
